@@ -28,7 +28,6 @@
 
 #include "xf86.h"
 #include "xf86_OSproc.h"
-#include "xf86_ansic.h"
 #include "xf86Priv.h"
 
 #include "xf86PciInfo.h"
@@ -38,6 +37,7 @@
 #include "GL/glxtokens.h"
 #include "sarea.h"
 
+#include "via.h"
 #include "via_driver.h"
 #include "via_drm.h"
 #include "via_dri.h"
@@ -66,8 +66,17 @@ extern void GlxSetVisualConfigs(
     void **configprivs
 );
 
+typedef struct {
+    int major;
+    int minor;
+    int patchlevel;
+} ViaDRMVersion;
+
 static char VIAKernelDriverName[] = "via";
 static char VIAClientDriverName[] = "unichrome";
+static const ViaDRMVersion drmExpected = {1, 3, 0};
+static const ViaDRMVersion drmCompat = {2, 0, 0};
+
 int test_alloc_FB(ScreenPtr pScreen, VIAPtr pVia, int Size);
 int test_alloc_AGP(ScreenPtr pScreen, VIAPtr pVia, int Size);
 static Bool VIAInitVisualConfigs(ScreenPtr pScreen);
@@ -160,13 +169,8 @@ VIADRIRingBufferInit(ScrnInfoPtr pScrn)
 
     if (pVia->agpEnable) {
 	drm_via_dma_init_t ringBufInit;
-	drmVersionPtr drmVer;
 
-	if (NULL == (drmVer = drmGetVersion(pVia->drmFD))) {
-	    return FALSE;
-	}
-
-	if (((drmVer->version_major <= 1) && (drmVer->version_minor <= 3))) {
+	if (((pVia->drmVerMajor == 1) && (pVia->drmVerMinor <= 3))) {
 	    return FALSE;
 	} 
 
@@ -333,11 +337,40 @@ static Bool VIADRIAgpInit(ScreenPtr pScreen, VIAPtr pVia)
 }
 static Bool VIADRIFBInit(ScreenPtr pScreen, VIAPtr pVia)
 {   
-    int FBSize = pVia->FBFreeEnd-pVia->FBFreeStart;
-    int FBOffset = pVia->FBFreeStart; 
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    int FBSize = pVia->driSize;
+    int FBOffset;
     VIADRIPtr pVIADRI = pVia->pDRIInfo->devPrivate;
+
+    if (FBSize < pVia->Bpl) {
+        xf86DrvMsg(pScreen->myNum, X_ERROR,
+		   "[drm] No DRM framebuffer heap available.\n");
+	xf86DrvMsg(pScreen->myNum, X_ERROR,
+		   "[drm] Please increase the frame buffer\n");
+	xf86DrvMsg(pScreen->myNum, X_ERROR,
+		   "[drm] memory area in BIOS. Disabling DRI.\n");
+	return FALSE;
+    }
+    if (FBSize < 3*(pScrn->virtualY * pVia->Bpl)) {
+	xf86DrvMsg(pScreen->myNum, X_WARNING,
+		   "[drm] The DRM Heap and Pixmap cache memory could be too small\n");
+	xf86DrvMsg(pScreen->myNum, X_WARNING,
+		   "[drm] for optimal performance. Please increase the frame buffer\n");
+	xf86DrvMsg(pScreen->myNum, X_WARNING,
+		   "[drm] memory area in BIOS.\n");
+    }
+
+    pVia->driOffScreenMem.pool = 0;
+    if (Success != viaOffScreenLinear(&pVia->driOffScreenMem, pScrn, FBSize)) {
+        xf86DrvMsg(pScreen->myNum, X_ERROR,
+		   "[drm] failed to allocate offscreen frame buffer area\n");
+	return FALSE;
+    }
+
+    FBOffset = pVia->driOffScreenMem.base;
+
     pVIADRI->fbOffset = FBOffset;
-    pVIADRI->fbSize = pVia->videoRambytes;
+    pVIADRI->fbSize = FBSize;
     
     {
 	drm_via_fb_t fb;
@@ -351,9 +384,7 @@ static Bool VIADRIFBInit(ScreenPtr pScreen, VIAPtr pVia)
 	    return FALSE;
 	} else {
 	    xf86DrvMsg(pScreen->myNum, X_INFO,
-		       "[drm] FBFreeStart= 0x%08x FBFreeEnd= 0x%08x "
-		       "FBSize= 0x%08x\n",
-		       pVia->FBFreeStart, pVia->FBFreeEnd, FBSize);
+		       "[drm] Using %d bytes for DRM memory heap.\n", FBSize);
 	    return TRUE;	
 	}   
     }
@@ -594,6 +625,7 @@ Bool VIADRIScreenInit(ScreenPtr pScreen)
     VIAPtr pVia = VIAPTR(pScrn);
     DRIInfoPtr pDRIInfo;
     VIADRIPtr pVIADRI;
+    drmVersionPtr drmVer;
 
     /* if symbols or version check fails, we still want this to be NULL */
     pVia->pDRIInfo = NULL;
@@ -601,7 +633,6 @@ Bool VIADRIScreenInit(ScreenPtr pScreen)
     /* Check that the GLX, DRI, and DRM modules have been loaded by testing
     * for canonical symbols in each module. */
     if (!xf86LoaderCheckSymbol("GlxSetVisualConfigs")) return FALSE;
-    if (!xf86LoaderCheckSymbol("DRIScreenInit"))       return FALSE;
     if (!xf86LoaderCheckSymbol("drmAvailable"))        return FALSE;
     if (!xf86LoaderCheckSymbol("DRIQueryVersion")) {
 	xf86DrvMsg(pScreen->myNum, X_ERROR,
@@ -636,9 +667,9 @@ Bool VIADRIScreenInit(ScreenPtr pScreen)
         ((pciConfigPtr)pVia->PciInfo->thisCard)->busnum,
         ((pciConfigPtr)pVia->PciInfo->thisCard)->devnum,
         ((pciConfigPtr)pVia->PciInfo->thisCard)->funcnum);
-    pDRIInfo->ddxDriverMajorVersion = VIA_DRI_VERSION_MAJOR;
-    pDRIInfo->ddxDriverMinorVersion = VIA_DRI_VERSION_MINOR;
-    pDRIInfo->ddxDriverPatchVersion = PATCHLEVEL;
+    pDRIInfo->ddxDriverMajorVersion = VIA_DRIDDX_VERSION_MAJOR;
+    pDRIInfo->ddxDriverMinorVersion = VIA_DRIDDX_VERSION_MINOR;
+    pDRIInfo->ddxDriverPatchVersion = VIA_DRIDDX_VERSION_PATCH;
     pDRIInfo->frameBufferPhysicalAddress = (pointer) pVia->FrameBufferBase;
     pDRIInfo->frameBufferSize = pVia->videoRambytes;  
   
@@ -662,7 +693,7 @@ Bool VIADRIScreenInit(ScreenPtr pScreen)
     /* For now the mapping works by using a fixed size defined
     * in the SAREA header
     */
-    if (sizeof(XF86DRISAREARec)+sizeof(VIASAREAPriv) > SAREA_MAX) {
+    if (sizeof(XF86DRISAREARec)+sizeof(drm_via_sarea_t) > SAREA_MAX) {
 	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Data does not fit in SAREA\n");
 	DRIDestroyInfoRec(pVia->pDRIInfo);
 	pVia->pDRIInfo = NULL;
@@ -698,6 +729,35 @@ Bool VIADRIScreenInit(ScreenPtr pScreen)
 	return FALSE;
     }
 
+    if (NULL == (drmVer = drmGetVersion(pVia->drmFD))) {
+	VIADRICloseScreen(pScreen);
+	return FALSE;
+    }
+    pVia->drmVerMajor = drmVer->version_major;
+    pVia->drmVerMinor = drmVer->version_minor;
+    pVia->drmVerPL = drmVer->version_patchlevel;
+
+    if ((drmVer->version_major < drmExpected.major) || 
+	(drmVer->version_major > drmCompat.major) ||
+	((drmVer->version_major == drmExpected.major ) && 
+	(drmVer->version_minor < drmExpected.minor))) {
+	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, 
+		   "[dri] Kernel drm is not compatible with this driver.\n"); 
+	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, 
+		   "[dri] Kernel drm version: %d.%d.%d "
+		   "and I can work with versions %d.%d.x - %d.x.x\n",
+		   drmVer->version_major,drmVer->version_minor,
+		   drmVer->version_patchlevel, drmExpected.major, 
+		   drmExpected.minor, drmCompat.major); 	
+	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, 
+		   "[dri] Please update either this 2D driver or your kernel DRM. "
+		   "Disabling DRI.\n");
+	drmFreeVersion(drmVer);
+	VIADRICloseScreen(pScreen);
+	return FALSE;
+    } 
+    drmFreeVersion(drmVer);
+
 	   
     if (!(VIAInitVisualConfigs(pScreen))) {
 	VIADRICloseScreen(pScreen);
@@ -711,10 +771,9 @@ Bool VIADRIScreenInit(ScreenPtr pScreen)
 	return FALSE;
     }
     pVIADRI->regs.size = VIA_MMIO_REGSIZE;
-    pVIADRI->regs.map = 0;
     pVIADRI->regs.handle = pVia->registerHandle;
     xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] mmio Registers = 0x%08lx\n",
-	pVIADRI->regs.handle);
+               (unsigned long) pVIADRI->regs.handle);
     
     pVIADRI->drixinerama = pVia->drixinerama;
 
@@ -741,8 +800,8 @@ VIADRICloseScreen(ScreenPtr pScreen)
 	drmAgpRelease(pVia->drmFD);
     }
 
-    
     DRICloseScreen(pScreen);
+    VIAFreeLinear(&pVia->driOffScreenMem);
     
     if (pVia->pDRIInfo) {
 	if ((pVIADRI = (VIADRIPtr) pVia->pDRIInfo->devPrivate)) {
@@ -817,10 +876,9 @@ VIADRIFinishScreenInit(ScreenPtr pScreen)
 
     /* set SAREA value */
     {
-	VIASAREAPriv *saPriv;
+	drm_via_sarea_t *saPriv;
 
-	saPriv=(VIASAREAPriv*)DRIGetSAREAPrivate(pScreen);
-	assert(saPriv);
+	saPriv=(drm_via_sarea_t *)DRIGetSAREAPrivate(pScreen);
 	memset(saPriv, 0, sizeof(*saPriv));
 	saPriv->ctxOwner = -1;
     }
@@ -883,7 +941,7 @@ static Bool VIADRIKernelInit(ScreenPtr pScreen, VIAPtr pVia)
     memset(&drmInfo, 0, sizeof(drm_via_init_t));
     drmInfo.func = VIA_INIT_MAP;
     drmInfo.sarea_priv_offset   = sizeof(XF86DRISAREARec);
-    drmInfo.fb_offset           = pVia->FrameBufferBase;
+    drmInfo.fb_offset           = pVia->frameBufferHandle;
     drmInfo.mmio_offset         = pVia->registerHandle;
     if (pVia->IsPCI)
 	drmInfo.agpAddr = (CARD32)NULL;
@@ -906,9 +964,14 @@ static Bool VIADRIMapInit(ScreenPtr pScreen, VIAPtr pVia)
 		  DRM_REGISTERS, flags, &pVia->registerHandle) < 0) {
 	return FALSE;
     }
+    xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] register handle = 0x%08lx\n",
+               (unsigned long) pVia->registerHandle);
+    if (drmAddMap(pVia->drmFD, pVia->FrameBufferBase, pVia->videoRambytes,
+		  DRM_FRAME_BUFFER, 0, &pVia->frameBufferHandle) < 0) {
+	return FALSE;
+    }
+    xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] framebuffer handle = 0x%08lx\n",
+               (unsigned long) pVia->frameBufferHandle);
     
-    xf86DrvMsg(pScreen->myNum, X_INFO,
-	"[drm] register handle = 0x%08lx\n", pVia->registerHandle);
-
     return TRUE;
 }
